#ifndef CondFormats_PhysicsToolsObjects_MVAComputer_h
#define CondFormats_PhysicsToolsObjects_MVAComputer_h
// -*- C++ -*-
//
// Package:     PhysicsToolsObjects
// Class  :     MVAComputer
//

//
// Author:  Christophe Saout <christophe.saout@cern.ch>
// Created:     Sat Apr 24 15:18 CEST 2007
// $Id: MVAComputer.h,v 1.7 2007/05/21 02:00:22 saout Exp $
//

#include <string>
#include <vector>
#include <map>

namespace PhysicsTools {
namespace Calibration {

// forward declarations

typedef std::pair<double, double> MinMax;

// helper classes

class BitSet {
    public:
  std::vector<unsigned char>  store;
  unsigned int      bitsInLast;
};

class Matrix {
    public:
  std::vector<double>    elements;
  unsigned int      rows;
  unsigned int      columns;
};

class PDF {
    public:
  std::vector<double>    distr;
  MinMax        range;
};

// configuration base classes

class VarProcessor {
    public:
  BitSet  inputVars;

  virtual ~VarProcessor() {}
  virtual std::string getInstanceName() const;
};

class Variable {
    public:
  std::string      name;
};

// variable processors

class ProcOptional : public VarProcessor {
    public:
  std::vector<double>    neutralPos;
  std::string getInstanceName() const { return "ProcOptional"; };
};

class ProcCount : public VarProcessor {};

class ProcClassed : public VarProcessor {
    public:
  unsigned int      nClasses;
  std::string getInstanceName() const { return "ProcClassed"; };
};

class ProcSplitter : public VarProcessor {
    public:
  unsigned int      nFirst;
    std::string getInstanceName() const { return "ProcSplitter"; };
};

class ProcNormalize : public VarProcessor {
    public:
  std::vector<PDF>    distr;
    std::string getInstanceName() const { return "ProcNormalize"; };
};

class ProcLikelihood : public VarProcessor {
    public:
  class SigBkg {
      public:
    PDF      signal;
    PDF      background;
    std::string getInstanceName() const { return "ProcLikelihood"; };
  };

  std::vector<SigBkg>    pdfs;
};

class ProcLinear : public VarProcessor {
    public:
  std::vector<double>    coeffs;
  double        offset;
  std::string getInstanceName() const { return "ProcLinear"; };
};

class ProcMultiply : public VarProcessor {
    public:
  typedef std::vector<unsigned int>  Config;
  std::string getInstanceName() const { return "ProcMultiply"; };

  unsigned int      in;
  std::vector<Config>    out;
};

class ProcMatrix : public VarProcessor {
    public:
  Matrix        matrix;
  std::string getInstanceName() const { return "ProcMatrix"; };
};

class ProcTMVA : public VarProcessor {
    public:
  std::string      method;
  std::vector<std::string>  variables;
  std::vector<unsigned char>  store;
  std::string getInstanceName() const { return "ProcTMVA"; };
};

class ProcMLP : public VarProcessor {
    public:
  typedef std::pair<double, std::vector<double> >  Neuron;
  typedef std::pair<std::vector<Neuron>, bool>  Layer;
  std::string getInstanceName() const { return "ProcMLP"; };

  std::vector<Layer>    layers;
};

// the discriminator computer

class MVAComputer {
    public:
  MVAComputer();
  virtual ~MVAComputer();
  std::vector<Variable>  inputSet; // the set of input tagging variables?
  virtual std::vector<VarProcessor*> getProcessors() const;
  unsigned int output; // the number of output variables?
  void        addProcessor(const VarProcessor *proc);
    private:
  std::vector<VarProcessor*>  processors_;
};

// this is a temporary hack used in RecoBTau until ESSources can be
// retrieved via label from the same record
class MVAComputerContainer {
    public:
  typedef std::pair<std::string, MVAComputer> Entry;

  MVAComputerContainer();
  virtual ~MVAComputerContainer() {}

  MVAComputer &add(const std::string &label);
  virtual const MVAComputer &find(const std::string &label) const;

  // cacheId stuff to detect changes
  typedef unsigned int CacheId;
  inline CacheId getCacheId() const { return cacheId; }
  inline bool changed(CacheId old) const { return old != cacheId; }

    private:
  std::vector<Entry>  entries;

  CacheId      cacheId;  // transient
};

} // namespace Calibration
} // namespace PhysicsTools

#endif // CondFormats_PhysicsToolsObjects_MVAComputer_h
