#include "RaveBase/Converters/interface/RaveToCmsObjects.h"
#include "RaveBase/RaveInterface/rave/Vector6D.h"
#include "DataFormats/GeometryVector/interface/GlobalPoint.h"
#include "DataFormats/GeometryVector/interface/GlobalVector.h"
#include "RaveTools/Converters/interface/MagneticFieldSingleton.h"
#include "RaveBase/Converters/interface/RaveToCLHEP.h"
#include "TrackingTools/TransientTrack/interface/TransientTrack.h"

#include <iostream>
#include <vector>

using namespace std;

GlobalTrajectoryParameters RaveToCmsObjects::convert ( const rave::Track & s ) const
{
  GlobalPoint pos ( s.state().x(), s.state().y(), s.state().z() );
  GlobalVector mom ( s.state().px(), s.state().py(), s.state().pz() );
  return GlobalTrajectoryParameters ( pos , mom, s.charge(), MagneticFieldSingleton::Instance() );
}

CartesianTrajectoryError RaveToCmsObjects::convert ( const rave::Covariance6D & c ) const
{
  AlgebraicSymMatrix s ( 6, 0 );
  s[0][0] = c.dxx();
  s[0][1] = c.dxy();
  s[0][2] = c.dxz();
  s[0][3] = c.dxpx();
  s[0][4] = c.dxpy();
  s[0][5] = c.dxpz();
  s[1][1] = c.dyy();
  s[1][2] = c.dyz();
  s[1][3] = c.dypx();
  s[1][4] = c.dypy();
  s[1][5] = c.dypz();
  s[2][2] = c.dzz();
  s[2][3] = c.dzpx();
  s[2][4] = c.dzpy();
  s[2][5] = c.dzpz();
  s[3][3] = c.dpxpx();
  s[3][4] = c.dpxpy();
  s[3][5] = c.dpxpz();
  s[4][4] = c.dpypy();
  s[4][5] = c.dpypz();
  s[5][5] = c.dpzpz();
  return CartesianTrajectoryError ( s );
}


GlobalPoint RaveToCmsObjects::convert ( const rave::Vector3D & v ) const
{
  return GlobalPoint ( v.x(), v.y(), v.z() );
}

AlgebraicVector RaveToCmsObjects::toAlgebraicVector ( const rave::Vector3D & v ) const
{
  AlgebraicVector av ( 3 );
  av[0] = v.x();
  av[1] = v.y();
  av[2] = v.z();
  return av;
}

GlobalError RaveToCmsObjects::convert ( const rave::Covariance3D & c ) const
{
  return GlobalError ( c.dxx(), c.dxy(), c.dyy(), c.dxz(), c.dyz(), c.dzz() );
}

TransientVertex RaveToCmsObjects::convert ( const rave::Vertex & v ) const
{
  GlobalPoint pos = convert ( v.position() );
  GlobalError err ( convert ( v.error() ) );
  vector < reco::TransientTrack > trks;
  vector< rave::Track > orig = v.tracks();
  for ( vector< rave::Track >::const_iterator i = orig.begin(); i != orig.end() ; ++i )
  {
    trks.push_back ( tTrack ( *i ) );
  }
  return TransientVertex ( pos, err, trks, v.chiSquared() );
}

reco::TransientTrack RaveToCmsObjects::tTrack ( const rave::Track & rt ) const
{
  AlgebraicSymMatrix cov6D = RaveToCLHEP().convert ( rt.error() );

  FreeTrajectoryState fts (
    GlobalTrajectoryParameters ( GlobalPoint ( rt.state().x(),
                                 rt.state().y(),
                                 rt.state().z() ),
                                 GlobalVector ( rt.state().px(),
                                                rt.state().py(),
                                                rt.state().pz() ),
                                 rt.charge(),
                                 MagneticFieldSingleton::Instance() ),
    CartesianTrajectoryError ( cov6D ) );
  GlobalPoint ref ( rt.state().x(), rt.state().y(), rt.state().z() );
  return reco::TransientTrack ( fts, ref, rt.id(), rt.chi2(), rt.ndof(), ( void * ) ( &rt ) );
}

#ifdef WITH_KINEMATICS

AlgebraicVector RaveToCmsObjects::toAlgebraicVector ( const rave::Vector4D & v ) const
{
  AlgebraicVector av ( 4 );
  av[0] = v.x();
  av[1] = v.y();
  av[2] = v.z();
  av[3] = v.m();
  return av;
}

::KinematicParameters RaveToCmsObjects::convert ( const rave::Vector7D & v ) const
{
  AlgebraicVector av ( 7 );
  av[0] = v.x();
  av[1] = v.y();
  av[2] = v.z();
  av[3] = v.px();
  av[4] = v.py();
  av[5] = v.pz();
  av[6] = v.m();
  return ::KinematicParameters ( av );
}

AlgebraicSymMatrix RaveToCmsObjects::toAlgebraicSymMatrix ( const rave::Covariance4D & c ) const
{
  AlgebraicSymMatrix s ( 4, 0 );
  s[0][0] = c.dxx();
  s[0][1] = c.dxy();
  s[0][2] = c.dxz();
  s[0][3] = c.dxm();
  s[1][1] = c.dyy();
  s[1][2] = c.dyz();
  s[1][3] = c.dym();
  s[2][2] = c.dzz();
  s[2][3] = c.dzm();
  s[3][3] = c.dmm();
  return s;
}

::KinematicParametersError RaveToCmsObjects::convert ( const rave::Covariance7D & c ) const
{
  AlgebraicSymMatrix s ( 7, 0 );
  s[0][0] = c.dxx();
  s[0][1] = c.dxy();
  s[0][2] = c.dxz();
  s[0][3] = c.dxpx();
  s[0][4] = c.dxpy();
  s[0][5] = c.dxpz();
  s[0][6] = c.dxm();
  s[1][1] = c.dyy();
  s[1][2] = c.dyz();
  s[1][3] = c.dypx();
  s[1][4] = c.dypy();
  s[1][5] = c.dypz();
  s[1][6] = c.dym();
  s[2][2] = c.dzz();
  s[2][3] = c.dzpx();
  s[2][4] = c.dzpy();
  s[2][5] = c.dzpz();
  s[2][6] = c.dzm();
  s[3][3] = c.dpxpx();
  s[3][4] = c.dpxpy();
  s[3][5] = c.dpxpz();
  s[3][6] = c.dpxm();
  s[4][4] = c.dpypy();
  s[4][5] = c.dpypz();
  s[4][6] = c.dpym();
  s[5][5] = c.dpzpz();
  s[5][6] = c.dpzm();
  s[6][6] = c.dmm();
  return ::KinematicParametersError ( s );
}

#endif

